import java.util.*;
import java.io.*;

public class MasterTag extends Tag {
	
	ArrayList m_labels = new ArrayList();
	
	public MasterTag( PropertyRecord record ) {
		super(null, null);
		setContext(record);
	}
	
	public boolean modifyRecord( PropertyRecord record ) {
		record.copyProperties(getContext());
		return false;
	}
	
	public void propogateRecommendations() {
		Statement stmt = this;
		while( stmt.getParent() != null && !(stmt.getParent() instanceof MasterTag) ) {
			stmt = stmt.getParent();
		}
		if( stmt.getParent() != null ) {
			MasterTag parent = (MasterTag)stmt.getParent();
			ArrayList open = getContext().openRecommendations();
			ArrayList close = getContext().closeRecommendations();
			ArrayList killList = new ArrayList();
			for( int i=0; i<open.size(); i++ ) {
				String tag = (String)open.get(i);
				if( parent.acceptsTag(tag, true) ) {
					parent.getContext().openRecommendations().add(tag);
					killList.add(tag);
				}
			}
			for( int i=0; i<killList.size(); i++ ) {
				open.remove(killList.get(i));
			}
			killList = new ArrayList();
			for( int i=0; i<close.size(); i++ ) {
				String tag = (String)close.get(i);
				if( parent.acceptsTag(tag, false) ) {
					parent.getContext().closeRecommendations().add(tag);
					killList.add(tag);
				}
			}
			for( int i=0; i<killList.size(); i++ ) {
				close.remove(killList.get(i));
			}
			parent.propogateRecommendations();
		}
		
	}
	
	public boolean acceptsTag( String tag, boolean open ) {
		boolean rc = false;
		if( tag.equals("B") ) {
			rc = getContext().isBold();
		} else if( tag.equals("EM") ) {
			// if the emphasis of the parent is different, return true
			if( getParent() == null ) {
				if( getContext().isEmphasis() ) {
					rc = true;
				} else if( getParent().getContext().isEmphasis() != getContext().isEmphasis() ) {
					rc = true;
				}
			}
		} else if( tag.equals("I") ) {
			rc = getContext().isItalics();
		} else if( tag.equals("PL") ) {
			if( !getContext().isStrong() &&
				!getContext().isEmphasis() &&
				!getContext().isItalics() &&
				!getContext().isBold() &&
				!getContext().isTypewriter() &&
				getContext().getUnderline() == 0 ) {
				rc = true;
			}
		} else if( tag.equals("S") ) {
			rc = getContext().isStrong();
		} else if( tag.equals("TT") ) {
			rc = getContext().isTypewriter();
		} else if( tag.equals("U") ) {
			ArrayList list = null;
			if( open ) {
				list = getContext().openRecommendations();
			} else {
				list = getContext().closeRecommendations();
			}
			int count = 0;
			for( int i=0; i<list.size(); i++ ) {
				if( list.get(i).equals("U") ) {
					count++;
				}
			}
			if( count < getContext().getUnderline() ) {
				rc = true;
			}
		} else if( tag.equals("0") ) {
			rc = (getContext().getSize() == 0);
		} else if( tag.equals("1") ) {
			rc = (getContext().getSize() == 1);
		} else if( tag.equals("2") ) {
			rc = (getContext().getSize() == 2);
		} else if( tag.equals("3") ) {
			rc = (getContext().getSize() == 3);
		} else if( tag.equals("4") ) {
			rc = (getContext().getSize() == 4);
		} else if( tag.equals("5") ) {
			rc = (getContext().getSize() == 5);
		} else if( tag.equals("6") ) {
			rc = (getContext().getSize() == 6);
		} else if( tag.equals("7") ) {
			rc = (getContext().getSize() == 7);
		} else if( tag.equals("8") ) {
			rc = (getContext().getSize() == 8);
		} else if( tag.equals("9") ) {
			rc = (getContext().getSize() == 9);
		} else if( tag.equals("r") ) {
			rc = tag.equals(getContext().getColorLabel());
		} else if( tag.equals("g") ) {
			rc = tag.equals(getContext().getColorLabel());
		} else if( tag.equals("b") ) {
			rc = tag.equals(getContext().getColorLabel());
		} else if( tag.equals("c") ) {
			rc = tag.equals(getContext().getColorLabel());
		} else if( tag.equals("m") ) {
			rc = tag.equals(getContext().getColorLabel());
		} else if( tag.equals("y") ) {
			rc = tag.equals(getContext().getColorLabel());
		} else if( tag.equals("k") ) {
			rc = tag.equals(getContext().getColorLabel());
		} else if( tag.equals("w") ) {
			rc = tag.equals(getContext().getColorLabel());
		}
		return rc;
	}
	
	public void writeString( PrintStream out ) {
		if( containsMismatachedTag() ) {
			System.err.print("");
		}
		ArrayList open = new ArrayList(getContext().openRecommendations());
		ArrayList close = new ArrayList(getContext().closeRecommendations());
		Stack stack = new Stack();
		int underlineCount = 0;
		for( int i=0; i<open.size(); i++ ) {
			// only print tags that have matching openings and closings here
			if( close.contains(open.get(i)) ) {
				close.remove(open.get(i));
				if( open.get(i).equals("U") ) {
					if( underlineCount < getContext().getUnderline() ) {
						underlineCount++;
						// make sure we haven't printed this too many times already
						stack.push(open.get(i));
						out.print("<"+open.get(i).toString()+">");
					}
				} else if( !stack.contains(open.get(i)) ) {
					// make sure we haven't printed this already
					stack.push(open.get(i));
					out.print("<"+open.get(i).toString()+">");
				}
			}
		}
		m_child.writeString(out);
		while( stack.size() > 0 ) {
			out.print("</"+stack.pop().toString()+">");
			
		}
	}	
}