import java.awt.*;
import java.util.*;

public class PropertyRecord implements Cloneable {
	private boolean m_bold = false;
	private boolean m_emphasis = false;
	private boolean m_italics = false;
	private boolean m_strong = false;
	private boolean m_typewriter = false;
	private int m_underline = 0;
	private int m_size = -1;
	private Color m_color = null;
	private String m_colorLabel = "";
	
	private Text m_text =  new Text("");
	
	public void setBold( boolean bold ) {
		m_bold = bold;
	}
	
	public void setEmphasis( boolean emphasis ) {
		m_emphasis = emphasis;
	}
	
	public void setItalics( boolean italics ) {
		m_italics = italics;
	}
	
	public void setStrong( boolean strong ) {
		m_strong = strong;
	}
	
	public void setType( boolean type ) {
		m_typewriter = type;
	}
	
	public void resetUnderline() {
		m_underline = 0;
	}
	
	public void incrementUnderline() {
		m_underline++;
		if( m_underline > 3 ) {
			m_underline = 3;
		}
	}
	
	public void setSize( int size ) {
		if( size > 9 ) {
			size = 9;
		}
		m_size = size;
	}
	
	public void setColor( Color color, String colorLabel ) {
		m_color = color;
		m_colorLabel = colorLabel;
	}
	
	public boolean isBold() {
		return m_bold;
	}
	
	public boolean isItalics() {
		return m_italics;
	}
	
	public boolean isEmphasis() {
		return m_emphasis;
	}
	
	public boolean isStrong() {
		return m_strong;
	}
	
	public boolean isTypewriter() {
		return m_typewriter;
	}
	
	public boolean isPlain() {
		boolean rc = true;
		rc = rc && (m_underline==0);
		rc = rc && !m_strong;
		rc = rc && !m_emphasis;
		rc = rc && !m_italics;
		rc = rc && !m_bold;
		rc = rc && !m_typewriter;
		return rc;
	}
	
	public void setText( Text text ) {
		m_text = text;
	}
	
	public Text getText() {
		return m_text;
	}
	
	public int getUnderline() {
		return m_underline;
	}
	
	public int getSize() {
		return m_size;
	}
	
	public Color getColor() {
		return m_color;
	}
	
	public String getColorLabel() {
		return m_colorLabel;
	}
		
	public void copyProperties( PropertyRecord record ) {
		m_bold = record.m_bold;
		m_emphasis = record.m_emphasis;
		m_italics = record.m_italics;
		m_strong = record.m_strong;
		m_typewriter = record.m_typewriter;
		m_underline = record.m_underline;
		m_size = record.m_size;
		m_color = record.m_color;
		m_colorLabel = record.m_colorLabel;
	}
	
	public boolean propertiesMatch( PropertyRecord record ) {
		boolean rc = true;
		
		rc = rc && (m_bold==record.m_bold);
		rc = rc && (m_emphasis==record.m_emphasis);
		rc = rc && (m_italics==record.m_italics);
		rc = rc && (m_strong==record.m_strong);
		rc = rc && (m_typewriter==record.m_typewriter);
		rc = rc && (m_underline==record.m_underline);
		rc = rc && (m_size==record.m_size);
		rc = rc && (m_color==record.m_color);
		rc = rc && (m_colorLabel.equals(record.m_colorLabel));
		
		return rc;
	}
	
	public boolean combine( PropertyRecord record ) {
		boolean rc = false;
		String text1 = getText().getString();
		String text2 = record.getText().getString();
		if( text2.length() == 0 ) {
			rc = true;
		} else if( 
			text2.length() == 1 && 
			Character.isWhitespace(text2.charAt(0)) && 
			!whitespaceDiffers( record ) ) {
			text1 += text2;
			getText().setString(text1);
			rc = true;
		} else if( 
			text1.length() == 1 &&
			Character.isWhitespace(text1.charAt(0)) &&
			!whitespaceDiffers( record ) ) {
			text1 += text2;
			getText().setString(text1);
			rc = true;
		} else if( propertiesMatch( record ) ) {
			text1 += text2;
			getText().setString(text1);
			rc = true;
		}
			
		return rc;
	}
	
	public boolean whitespaceDiffers( PropertyRecord record ) {
		boolean rc = true;
		if( ( getUnderline() == record.getUnderline() ) && 
			( getSize() == record.getSize() ) && 
			( ( getColor() == record.getColor() ) || (getUnderline() == 0) ) ) {
			rc = false;
		}
		return rc;
	}
	
	public PropertyRecord createCopy() {
		PropertyRecord rc = null;
		try {
			rc = (PropertyRecord)this.clone();
		} catch( CloneNotSupportedException e ) {
			System.err.println(e);
		}
		return rc;
	}
	
	protected Object clone() throws CloneNotSupportedException {
		PropertyRecord rc = (PropertyRecord)super.clone();
		rc.m_open = new ArrayList();
		rc.m_close = new ArrayList();
		return rc;
	}
	
	public void buildTagRecommendations(PropertyRecord prev, PropertyRecord next) {
		m_open.clear();
		m_close.clear();
		buildDifferences( prev, m_open );
		buildDifferences( next, m_close );		
	}
	
	ArrayList m_open = new ArrayList();
	public ArrayList openRecommendations() {
		return m_open;
	}
	
	ArrayList m_close = new ArrayList();
	public ArrayList closeRecommendations() {
		return m_close;
	}
	
	private void buildDifferences( PropertyRecord record, ArrayList list ) {
		if( m_color != record.m_color ) {
			list.add(m_colorLabel);
		}
		if( m_size != record.m_size ) {
			list.add(""+m_size);
		}
		if( (m_underline - record.m_underline) == 1 ) {
			list.add("U");
		}
		if( (m_underline - record.m_underline) == 2 ) {
			list.add("U");
			list.add("U");
		}
		if( (m_underline - record.m_underline) == 3 ) {
			list.add("U");
			list.add("U");
			list.add("U");
		}
		if( m_typewriter && !record.m_typewriter ) {
			list.add("TT");
		}
		if( m_strong && !record.m_strong ) {
			list.add("S");
		}
		if( m_italics && !record.m_italics ) {
			list.add("I");
		}
		if( m_emphasis && !record.m_emphasis ) {
			list.add("EM");
		}
		if( m_bold && !record.m_bold ) {
			list.add("B");
		}
	}
	
	public String openDifference( PropertyRecord record ) {
		StringBuffer rc = new StringBuffer();
		if( m_color != record.m_color ) {
			rc.append("<"+m_colorLabel+">");
		}
		if( m_size != record.m_size ) {
			rc.append("<"+m_size+">");
		}
		if( (m_underline - record.m_underline) == 1 ) {
			rc.append("<U>");
		}
		if( (m_underline - record.m_underline) == 2 ) {
			rc.append("<U><U>");
		}
		if( (m_underline - record.m_underline) == 3 ) {
			rc.append("<U><U><U>");
		}
		if( m_typewriter && !record.m_typewriter ) {
			rc.append("<TT>");
		}
		if( m_strong && !record.m_strong ) {
			rc.append("<S>");
		}
		if( m_italics && !record.m_italics ) {
			rc.append("<I>");
		}
		if( m_emphasis && !record.m_emphasis ) {
			rc.append("<EM>");
		}
		if( m_bold && !record.m_bold ) {
			rc.append("<B>");
		}
		
		return rc.toString();
	}
	
	public String closeDifference( PropertyRecord record ) {
		StringBuffer rc = new StringBuffer();
		if( m_bold && !record.m_bold ) {
			rc.append("</B>");
		}
		if( m_emphasis && !record.m_emphasis ) {
			rc.append("</EM>");
		}
		if( m_italics && !record.m_italics ) {
			rc.append("</I>");
		}
		if( m_strong && !record.m_strong ) {
			rc.append("</S>");
		}
		if( m_typewriter && !record.m_typewriter ) {
			rc.append("</TT>");
		}
		if( (m_underline - record.m_underline) == 1 ) {
			rc.append("</U>");
		}
		if( (m_underline - record.m_underline) == 2 ) {
			rc.append("</U></U>");
		}
		if( (m_underline - record.m_underline) == 3 ) {
			rc.append("</U></U></U>");
		}
		if( m_size != record.m_size ) {
			rc.append("</"+m_size+">");
		}
		if( m_color != record.m_color ) {
			rc.append("</"+m_colorLabel+">");
		}
		return rc.toString();
	}
}