import java.io.*;
import java.util.*;

public class SMLMeaning {
	ArrayList m_recordList = new ArrayList();
	
	public void buildMeaning( Statement sml ) {
		sml.setContext(new PropertyRecord());
		Stack stack = new Stack();
		stack.push( sml );
		while( stack.size() > 0 ) {
			Statement stmt = (Statement)stack.pop();
			PropertyRecord context = stmt.getContext();
			if( stmt.modifyRecord( context ) ) {
				m_recordList.add(context);
			}
			if( !(stmt instanceof Tag ) ) {
//				stmt.installMasterTag( context );				
			}
			stmt.setChildContexts( context );
			stmt.pushChildren( stack );
		}
		
		boolean simplify = true;
		while( simplify ) {
			simplify = false;
			simplify = simplify || combineRecords();
			simplify = simplify || removeWhitespace();
		}
//		buildTagRecommendations();
	}
		
	private boolean combineRecords() {
		boolean rc = false;
		ArrayList killList = new ArrayList();
		PropertyRecord prev = null;
		Iterator iter = m_recordList.iterator();
		while( iter.hasNext() ) {
			PropertyRecord record = (PropertyRecord)iter.next();
			if( prev != null && 
				record != null && 
				prev.combine(record) ) {
				killList.add(record);
				record.getText().removeText();
			} else {
				prev = record;
			}
		}
		for( int i=0; i<killList.size(); i++ ) {
			m_recordList.remove(killList.get(i));
		}
		return rc;
	}
	
	private boolean removeWhitespace() {
		boolean rc = false;
		Object[] objArray = m_recordList.toArray();
		PropertyRecord[] records = new PropertyRecord[objArray.length+2];
		records[0] = new PropertyRecord();
		records[ records.length-1 ] = new PropertyRecord();
		System.arraycopy( objArray, 0, records, 1, objArray.length );
		
		for( int i=1; i<records.length-1; i++ ) {
			if( !records[i].isTypewriter() ) {
				// remove extra whitespace from the string
				StringBuffer newText = new StringBuffer();
				String text = records[i].getText().getString();
				for( int j=0; j< text.length(); j++ ) {
					char c = text.charAt(j);
					if( Character.isWhitespace(c) ) {
						// read through the rest of the whitespace
						while( ( (j+1) < text.length() ) && ( Character.isWhitespace( text.charAt(j+1) ) ) ) {
							rc = true;
							j++;
						}
					}
					newText.append(c);
				}
				records[i].getText().setString(newText.toString());
				// if properties do not affect whitespace,
				// and the previous record ends in whitespace,
				// and we begin in whitespace, remove the whitespace
				// from our front
				if( ( records[i].getUnderline() == records[i-1].getUnderline() ) && 
					( records[i].getSize() == records[i-1].getSize() ) && 
					( ( records[i].getColor() == records[i-1].getColor() ) || (records[i].getUnderline() == 0) ) ) {
					String prevText = records[i-1].getText().getString();
					text = records[i].getText().getString();
					if( prevText.length() > 0 &&
						Character.isWhitespace(prevText.charAt( prevText.length()-1)) && 
						text.length() > 0 &&
						Character.isWhitespace(text.charAt(0)) ) {
						text = text.substring(1);
						rc = true;
					}
					records[i].getText().setString(text);
				}
				
			}
		}
		return rc;
	}
	
	public void buildTagRecommendations() {
		Object[] objArray = m_recordList.toArray();
		PropertyRecord[] records = new PropertyRecord[objArray.length+2];
		records[0] = new PropertyRecord();
		records[ records.length-1 ] = new PropertyRecord();
		System.arraycopy( objArray, 0, records, 1, objArray.length );
		
		for( int i=1; i<records.length-1; i++ ) {
			records[i].buildTagRecommendations(records[i-1], records[i+1]);
		}
		
		return;
	}
	
	public void writeString(PrintStream out) {
		Object[] objArray = m_recordList.toArray();
		PropertyRecord[] records = new PropertyRecord[objArray.length+2];
		records[0] = new PropertyRecord();
		records[ records.length-1 ] = new PropertyRecord();
		System.arraycopy( objArray, 0, records, 1, objArray.length );
		
		for( int i=1; i<records.length-1; i++ ) {
			// compare current record to previous record
			
			out.print( records[i].openDifference( records[i-1] ) );
			// output the text
			out.print( records[i].getText() );
			// and compare to next record
			out.print( records[i].closeDifference( records[i+1] ) );
		}
		
		return;
	}
}