import java.io.*;
import java.util.*;

public abstract class Statement {
	protected PropertyRecord m_record = null;
	protected Statement m_parent = null;
	public final void setContext( PropertyRecord record ) {
		m_record = record;
	}
	
	public final PropertyRecord getContext() {
		return m_record;
	}
	
	public final void installMasterTag( PropertyRecord record ) {
		Statement tagRoot = this;
		while( ( tagRoot.getParent() != null ) && ( tagRoot.getParent() instanceof Tag ) ) {
			tagRoot = tagRoot.getParent();
		}
		if( tagRoot.getParent() != null ) {
			MasterTag tag = new MasterTag(record);
			tagRoot.getParent().addChild(tag);
			tagRoot.getParent().removeChild(tagRoot);
			this.getParent().removeChild(this);
			tag.addChild(this);
		}
	}
	
	public void processRecommendations() {
		Stack stack = new Stack();
		stack.push(this);
		while( stack.size() > 0 ) {
			Statement stmt = (Statement)stack.pop();
			if( stmt instanceof Text ) {
				MasterTag tag = (MasterTag)stmt.getParent();
				tag.propogateRecommendations();
			} else {
				stmt.pushChildren(stack);
			}
		}
	}
	
	public boolean containsMismatachedTag() {
		ArrayList open = m_record.openRecommendations();
		ArrayList close = m_record.closeRecommendations();
		Collections.sort(open);
		Collections.sort(close);
		return !open.equals(close);
	}
	
	protected final void setParent( Statement parent ) {
		m_parent = parent;
	}
	
	protected final Statement getParent() {
		return m_parent;
	}
	
	public abstract boolean modifyRecord( PropertyRecord record );
	public abstract void setChildContexts( PropertyRecord context );
	public abstract void pushChildren( java.util.Stack stack );
	public abstract void writeString( PrintStream out );	
	protected abstract void addChild( Statement child );
	protected abstract void removeChild( Statement child );
}